@@ -7,7 +7,7 @@ use Net::OpenStack::Swift::Util qw/uri_escape uri_unescape debugf/;
 use Net::OpenStack::Swift::InnerKeystone;
 use namespace::clean -except => 'meta';
 
-our $VERSION = "0.02";
+our $VERSION = "0.03";
 
 
 has auth_version => (is => 'rw', required => 1, default => sub {"2.0"}); 
@@ -17,7 +17,6 @@ has password     => (is => 'rw', required => 1);
 has tenant_name  => (is => 'rw');
 has storage_url  => (is => 'rw');
 has token        => (is => 'rw');
-#has verify_ssl   => (is => 'ro', default => sub {! $ENV{OSCOMPUTE_INSECURE}});
 has agent => (
     is      => 'rw',
     lazy    => 1,
@@ -39,6 +38,7 @@ sub _request {
         special_headers => $args->{special_headers},
         headers         => $args->{header},
         write_code      => $args->{write_code},
+        write_file      => $args->{write_file},
         content         => $args->{content},
     );
     return $res;
@@ -160,7 +160,6 @@ sub post_account {
     debugf("post_account() response body %s",    $res->content);
     my %headers = @headers;
     return \%headers;
-
 }
 
 sub get_container {
@@ -321,14 +320,15 @@ sub delete_container {
 sub get_object {
     my $self = shift;
     my $rule = Data::Validator->new(
-        url            => { isa => 'Str', default => $self->storage_url},
+        url            => { isa => 'Str', default => $self->storage_url },
         token          => { isa => 'Str', default => $self->token },
         container_name => { isa => 'Str'},
         object_name    => { isa => 'Str'},
-        write_code     => { isa => 'CodeRef'},
+        write_file     => { isa => 'FileHandle', xor => [qw(write_code)] },
+        write_code     => { isa => 'CodeRef' },
     );
     my $args = $rule->validate(@_);
-
+    
     my $request_header = ['X-Auth-Token' => $args->{token}];
     my $request_url    = sprintf "%s/%s/%s", $args->{url}, 
         uri_escape($args->{container_name}), 
@@ -338,11 +338,19 @@ sub get_object {
     debugf("get_object() request special headers: %s", $request_url);
     debugf("get_object() request url: %s", $request_url);
 
-    my $res = $self->_request({
+    my $request_params = {
         method => 'GET', url => $request_url, header => $request_header, 
         special_headers => \%special_headers,
-        write_code      => $args->{write_code}
-    });
+        write_code      => undef,
+        write_file      => undef,
+    };
+    if (exists $args->{write_code}) {
+        $request_params->{write_code} = $args->{write_code};
+    }
+    if (exists $args->{write_file}) {
+        $request_params->{write_file} = $args->{write_file};
+    }
+    my $res = $self->_request($request_params);
 
     croak "Object GET failed: ".$res->status_line unless $res->is_success;
     my @headers = $res->headers->flatten();
@@ -392,8 +400,10 @@ sub put_object {
         container_name => { isa => 'Str'},
         object_name    => { isa => 'Str'},
         content        => { isa => 'Str|FileHandle'},
-        content_length => { isa => 'Int'},
+        content_length => { isa => 'Int', default => sub { 0 } },
         content_type   => { isa => 'Str', default => 'application/octet-stream'},
+        etag           => { isa => 'Str', default => undef },
+        query_string   => { isa => 'Str', default => sub {''} },
     );
     my $args = $rule->validate(@_);
 
@@ -402,9 +412,16 @@ sub put_object {
         'Content-Length' => $args->{content_length}, 
         'Content-Type'   => $args->{content_type}, 
     ];
+    if ($args->{etag}) {
+        push @{$request_header}, ('ETag' => $args->{etag});
+    }
+ 
     my $request_url = sprintf "%s/%s/%s", $args->{url}, 
         uri_escape($args->{container_name}), 
         uri_escape($args->{object_name}); 
+    if ($args->{query_string}) {
+        $request_url .= '?'.$args->{query_string}; 
+    }
     debugf("put_object() request header %s", $request_header);
     debugf("put_object() request url: %s", $request_url);
 
@@ -602,27 +619,25 @@ Show account metadata.
 =head2 post_account
 
 Create, update, or delete account metadata.
-
+    
 =head2 get_container
 
 Show container details and list objects.
 
+    my ($headers, $containers) = $sw->get_container(container_name => 'container1');
+
 =head2 head_container
 
 Show container metadata.
 
+    my $headers = $sw->head_container(container_name => 'container1');
+
 =head2 put_container
 
 Create container.
 
     my $headers = $sw->put_container(container_name => 'container1')
 
-=over
-
-=item container_name
-
-=back
-
 =head2 post_container
 
 Create, update, or delete container metadata.
@@ -630,11 +645,18 @@ Create, update, or delete container metadata.
 =head2 delete_container
 
 Delete container.
+
+    my $headers = $sw->delete_container(container_name => 'container1');
     
 =head2 get_object
 
 Get object content and metadata.
 
+    open my $fh, ">>:raw", "hoge.jpeg" or die $!; 
+    my $etag = $sw->get_object(container_name => 'container_name1', object_name => 'masakystjpeg', 
+        write_file => $fh,
+    );
+    # or chunked
     open my $fh, ">>:raw", "hoge.jpeg" or die $!; 
     my $etag = $sw->get_object(container_name => 'container1', object_name => 'hoge.jpeg', 
         write_code => sub {
@@ -650,9 +672,13 @@ Get object content and metadata.
 
 =item object_name
 
-=item write_code
+=item write_file: FileHandle
 
-Code reference
+the response content will be saved here instead of in the response object.
+
+=item write_code: Code reference
+
+the response content will be called for each chunk of the response content.
 
 =back
 
@@ -673,9 +699,14 @@ Create or replace object.
 
 =over
 
-=item content
+=item content: Str|FileHandle
 
-String or FileHandle
+=item content_length: Int
+
+=item content_type: Str
+
+Optional.
+default 'application/octet-stream'
 
 =back
 
